#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-now() - The Aura Engine Team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Domain models for scheduling.

These domain models are defined:
    - Timeslot: A timetable item.
    - Show: Information on the show related to a timeslot.
    - Episode: Information on the episode related to a timeslot
    - Playlists: Container for all types of playlists assignable to a timeslot.
    - Playlist: Definition of what should be broadcasted for one timeslot.
    - PlaylistType: The kind of playlist it is; either assigned to timeslot, schedule or show.
    - PlaylistItem: An element of the playlist, referencing media sources and their play state.
    - PlayState: State information for a given playlist item.
"""

from __future__ import annotations

import enum
from typing import NamedTuple

from aura_engine import resources
from aura_engine.base.lang import synchronized
from aura_engine.base.utils import SimpleUtil as SU
from aura_engine.core.channels import GenericChannel


class PlaylistType(enum.Enum):
    """
    Playlist type defines on which level the playlist is assigned to the timeslot.

    Attributes:
        TIMESLOT (str): The playlist is assigned to the timeslot itself.
        SCHEDULE (str): The playlist is assigned to the schedule.
        SHOW (str): The playlist is assigned to the show.

    """

    TIMESLOT = "timeslot"
    SCHEDULE = "schedule"
    SHOW = "show"


class Playlist:
    """
    The playlist assigned to a timeslot.

    Attributes:
        id (int): Playlist ID as used in Steering and Tank.
        timeslot (Timeslot): Back reference to the timeslot.
        type (PlaylistType): Indicating type of timeslot playlist.
        items (PlaylistItem): List of items.

    """

    id: int
    timeslot: Timeslot
    type: PlaylistType
    desc: str
    shuffle: bool
    items: list[PlaylistItem]

    def __init__(self, id: int, desc: str, shuffle: bool = False):
        """
        Initialize.

        Args:
            id (int): The playlist ID.
            desc: (str): Description for the playlist.
            shuffle (bool): If the playlist should be played in random order.
        """
        self.id = id
        self.desc = desc
        self.shuffle = shuffle
        self.items = []
        self.timeslot = None

    def get_id(self) -> int:
        """
        Get the playlist ID.

        Returns:
            int: ID of the playlist.
        """
        return self.id

    def get_timeslot(self) -> Timeslot:
        """
        Set the timeslot.

        Returns:
            (Timeslot): The timeslot of the playlist.
        """
        return self.timeslot

    def set_timeslot(self, timeslot: Timeslot):
        """
        Set the timeslot.

        Args:
            Timeslot: The timeslot of the playlist.
        """
        self.timeslot = timeslot

    def get_type(self) -> PlaylistType:
        """
        Get the playlist type.

        Returns:
            PlaylistType: Type of the playlist.
        """
        return self.type

    def get_items(self) -> list[PlaylistItem]:
        """
        Get the items of the playlist.

        Returns:
            list[PlaylistItem]: All playlist items.
        """
        return self.items

    def get_duration(self) -> float:
        """
        Return the total length of the playlist in seconds.

        Returns:
            float: Length in seconds.

        """
        return sum(item.duration for item in self.items)

    def get_description(self) -> str:
        """
        Get playlist description.

        Returns:
            str: The description.
        """
        return self.desc

    def do_shuffle(self) -> bool:
        """
        Returns true if the playlist should be played in random order.

        Returns:
            bool: `False` if playout mode is linear.

        """
        return self.shuffle

    @synchronized
    def add(self, item: PlaylistItem) -> None:
        """
        Add an item to the playlist.

        It also adds the back-reference to the playlist and performs linking to the previous
        playlist item.

        Args:
            item (PlaylistItem): The playlist item to append.

        """
        prev = None
        if self.items:
            prev = self.items[-1]
            item.prev = prev
            item.prev.next = item
        item.set_playlist(self)
        self.items.append(item)

    @synchronized
    def update_playlist(self, playlist: Playlist):
        """
        Update the playlist data with the one passed.

        It is important to avoid overwriting any local references, as they might be already in use
        of scheduling threads.

        Args:
            playlist (Playlist): Playlist with fresh data.
        """
        self.id = playlist.id
        self.desc = playlist.desc
        self.shuffle = playlist.shuffle
        self.items = playlist.get_items()
        item: PlaylistItem
        for item in self.get_items():
            item.set_playlist(self)

    def __str__(self) -> str:
        """
        String representation.

        Returns:
            str: Playlist as String.

        """
        duration = "{:.2f}".format(self.get_duration())
        s = f"ID#{self.id} | {duration} seconds"
        if self.desc:
            s += f" ({self.desc})"
        return s


class Playlists:
    """
    Container to hold all timeslot assignable types of playlists.

    Attributes:
        timeslot (Playlist): The playlist assigned on timeslot level.
        schedule (Playlist): The playlist assigned on schedule level.
        show (Playlist): The playlist assigned on show level.

    """

    timeslot: Playlist
    schedule: Playlist
    show: Playlist

    def __init__(self, timeslot: Playlist, schedule: Playlist, show: Playlist):
        """
        Initialize.

        Args:
            timeslot (Playlist): The playlist assigned on timeslot level.
            schedule (Playlist): The playlist assigned on schedule level.
            show (Playlist): The playlist assigned on show level.
        """
        self.timeslot = timeslot
        self.schedule = schedule
        self.show = show

    def get(self, type: PlaylistType) -> Playlist:
        """
        Retrieve a playlist by type.

        Args:
            type (PlaylistType): The type of playlist.

        Returns:
            Playlist: The playlist.
        """
        return getattr(self, type.value)

    def set(self, type: PlaylistType, playlist: Playlist):
        """
        Retrieve a playlist by type.

        Args:
            type (PlaylistType): The type of playlist.
            playlist (Playlist): The playlist to update.

        """
        setattr(self, type.value, playlist)
        if playlist:
            playlist.type = type


Show = NamedTuple(
    "Show",
    [
        ("id", str),
        ("name", str),
        ("type", str),
        ("cat", str),
        ("funding_cat", str),
        ("music_focus", str),
    ],
)

Episode = NamedTuple(
    "Episode",
    [
        ("title", str),
        ("hosts", list[str]),
        ("languages", list[str]),
        ("topics", list[str]),
        ("memo", str),
    ],
)


class Timeslot:
    """
    A timeslot in the programme calendar.

    Beside references to playlists, it holds show and episode metadata.
    Note, some episodes values are inherited from the show,
    some others are overridden by the episode.

    Attributes:
        id (int): Timeslot ID as used in Steering and Tank
        repetition_id (int): In case of a re-broadcast it holds a reference to another timeslot ID.
        start: beginning of timeslot as UNIX epoch.
        end: ending of timeslot as UNIX epoch.

        show.id (int): Show ID as used in Steering and Tank.
        show.name (str): Name of the show.
        show.type (str): The type.
        show.cat (str): The category.
        show.funding_cat (str): The funding category.
        show.music_focus (str): The primary music focus.

        episode.title (str): Title of the episode.
        episode.hosts (list[str]): List of hosts.
        episode.languages (list[str]): List of languages.
        episode.topics (list[str]): List of topics.
        episode.memo: Some notes for the (internal, non-public use only).

        playlists.timeslot (Playlist): Playlist assigned directly to the timeslot.
        playlists.schedule (Playlist): Playlist assigned to the schedule.
        paylists.show (Playlist): Playlist assigned to the show.

    """

    id: int
    repetition_id = int
    start: int
    end: int

    show: Show
    episode: Episode
    playlists: Playlists

    def __init__(
        self, id: int, repetition_id: int, start: float, end: float, show: Show, episode: Episode
    ):
        """
        Initialize.

        Args:
            id (int): ID of the timeslot.
            repetition_id (int): ID of the timeslot this timeslot is a repetition of. None if n/a.
            start (int): Start of the timeslot as UNIX timestamp in seconds.
            end (int): Start of the timeslot as UNIX timestamp in seconds.
            show (Show): Show assigned to the timeslot
            episode (Episode): Episode information representing this broadcast.
        """
        self.id = id
        self.repetition_id = repetition_id
        self.start = start
        self.end = end
        self.show = show
        self.episode = episode
        self.playlists = Playlists(None, None, None)

    def get_id(self) -> int:
        """
        Get ID.

        Returns:
            int: The ID.
        """
        return self.id

    def get_repetition_id(self) -> int:
        """
        Get ID of the timeslot this timeslot is a repetition of.

        Returns:
            int: Repetition timeslot ID. None if it is not a repeating timeslot.
        """
        return self.repetition_id

    def set_playlists(self, pl_timeslot: Playlist, pl_schedule: Playlist, pl_show: Playlist):
        """
        Set all playlists and their respective types at once.

        Args:
            pl_timeslot (Playlist): The timeslot playlist.
            pl_schedule (Playlist): The schedule playlist.
            pl_show (Playlist): The show playlist.

        """
        if pl_timeslot:
            pl_timeslot.type = PlaylistType.TIMESLOT
            pl_timeslot.set_timeslot(self)
        if pl_schedule:
            pl_schedule.type = PlaylistType.SCHEDULE
            pl_schedule.set_timeslot(self)
        if pl_show:
            pl_show.type = PlaylistType.SHOW
            pl_show.set_timeslot(self)

        self.playlists = Playlists(timeslot=pl_timeslot, schedule=pl_schedule, show=pl_show)

    @synchronized
    def update(self, timeslot: Timeslot):
        """
        Update the timeslot with a newer version.

        Args:
            timeslot (Timeslot): The new timeslot.

        """
        self.show = timeslot.show
        self.episode = timeslot.episode
        self.id = timeslot.id
        self.repetition_id = timeslot.repetition_id
        self.update_playlist(PlaylistType.TIMESLOT, timeslot.playlists.timeslot)
        self.update_playlist(PlaylistType.SCHEDULE, timeslot.playlists.schedule)
        self.update_playlist(PlaylistType.SHOW, timeslot.playlists.show)

    @synchronized
    def update_playlist(self, type: PlaylistType, playlist: Playlist):
        """
        Update the playlist data with the one passed.

        It is important to avoid overwriting any local references, as they might be already in use
        of scheduling threads.

        Args:
            playlist (Playlist): Playlist with fresh data.
        """
        if self.playlists.get(type) and playlist:
            pl: Playlist = self.playlists.get(type)
            pl.update_playlist(playlist)
        else:
            self.playlists.set(type, playlist)

    def get_current_playlist(self) -> Playlist:
        """
        Retrieve the playlist to be scheduled.

        The playlist to be retrieved is checked in this order:

            1. Timeslot playlist
            2. Schedule playlist
            3. Show playlist

        Returns:
            (Playlist): The playlist currently valid for being scheduled.

        """
        playlist = self.playlists.timeslot
        if not playlist:
            playlist = self.playlists.schedule
            if not playlist:
                playlist = self.playlists.show

        return playlist

    def get_start(self) -> float:
        """
        Get start of timeslot.

        Returns:
            float: start as UNIX epoch timestamp.
        """
        return self.start

    def get_end(self) -> float:
        """
        Get end of timeslot.

        Returns:
            float: end as UNIX epoch timestamp.
        """
        return self.end

    def get_duration(self) -> float:
        """
        Get duration of timeslot in seconds.

        Note, this is not the duration of some actual playlist.

        Returns:
            float: timeslot length in seconds.
        """
        return self.get_end() - self.get_start()

    def get_show(self) -> Show:
        """
        Get show associated with the timeslot.

        Returns:
            Show: The show.
        """
        return self.show

    def get_episode(self) -> Episode:
        """
        Get episode associated with the timeslot.

        Returns:
            Episode: The episode.
        """
        return self.episode

    def __str__(self) -> str:
        """
        String representation.

        Returns:
            str: Timeslot as String.

        """
        start = SU.fmt_time(self.start)
        end = SU.fmt_time(self.end)
        s = f"ID#{self.id}: {start} - {end}"
        if self.show:
            s += f" [Show#{self.show.id}: {self.show.name}]"
        return s


class PlayState:
    """
    Play-out state information on a queued or an already playing playlist item.

    It holds info on play-state, the chosen channel and the actual start time.
    """

    class PlayStateType(str, enum.Enum):
        """
        Play-state of a playlist item.
        """

        UNKNOWN = "unknown"
        LOADING = "loading"
        READY = "ready"
        PLAYING = "playing"
        DONE = "done"
        FAILED = "failed"

    state: PlayStateType
    play_start: float
    channel: GenericChannel

    def __init__(self):
        """
        Initialize the play info.
        """
        self.state = None
        self.play_start = None
        self.channel = None

    def is_loading(self) -> bool:
        """
        Check if item is loading.
        """
        return self.state == self.PlayStateType.LOADING

    def set_loading(self, channel: GenericChannel):
        """
        Set item state to loading and stores the used channel.

        Args:
            channel (GenericChannel): The channel where the item is loaded into.
        """
        self.state = self.PlayStateType.LOADING
        self.channel = channel

    def is_ready(self) -> bool:
        """
        Check if item is ready to play.
        """
        return self.state == self.PlayStateType.READY

    def set_ready(self):
        """
        Set item status as ready to play.
        """
        self.state = self.PlayStateType.READY

    def is_playing(self) -> bool:
        """
        Check if item is ready playing.
        """
        return self.state == self.PlayStateType.PLAYING

    def set_playing(self):
        """
        Set item status as playing and assigns current timestamp as start time.

        Note the start time is not guaranteed to match the exact play-out time. The actual play-out
        start time can only be gathered on the site of Liquidsoap.
        """
        self.state = self.PlayStateType.PLAYING
        self.play_start = SU.timestamp()

    def get_play_start(self) -> float:
        """
        Get the time where the item actually started playing.

        Important things to note:
            - Do not mix this up with the planned start time (See `PlaylistItem.get_start_time()`).
            - Some of the play time is set in Engine (line sources), some other natively by
              Liquidsoap (audio files).

        Returns:
            float: Play start time as timestamp.
        """
        return self.play_start

    def is_done(self) -> bool:
        """
        Check if item is done playing.
        """
        return self.state == self.PlayStateType.DONE

    def set_done(self):
        """
        Set item status as done playing.
        """
        self.state = self.PlayStateType.DONE


class PlaylistItem:
    """
    An item of some playlist.

    It holds an reference to the playlist and a double linked list to the previous and next item.

    Attributes:
        prev (PlaylistItem): The previous playlist item.
        next (PlaylistItem): The next playlist item.

        playlist (str): Back reference to playlist.
        source (str): URI referencing the audio source.
        duration (float): Duration in seconds.
        volume (int): Value from 0-100% indicating loudness.

        play (PlayState): Information on play-out state, start time and used channel.
        play_queue_state (QueueState): Info how the item behaves within the timeslot.
    """

    prev: PlaylistItem
    next: PlaylistItem

    playlist: Playlist
    source: str
    duration: float
    volume: int
    play: PlayState

    Metadata = NamedTuple("Metadata", [("artist", str), ("album", str), ("title", str)])
    metadata: Metadata

    class QueueState(enum.Enum):
        """
        Types of playlist item behaviours.

        Items are either fully played, cut or not played at all, because they are past the
        timeslot.

        It is only used for displaying in the list of queued media in the logs, compare
        the logic of `TimetableRenderer`.

        TODO Think about refactoring this out of the core business logic.
        """

        OKAY = "ok"
        CUT = "cut"
        PAST_TIMESLOT = "pt"

    queue_state: QueueState

    def __init__(self, source: str, duration: float, volume: int, metadata: Metadata):
        """
        Initialize the playlist item.

        Args:
            source (str): URI referencing the media source.
            duration (float): Duration in seconds.
            volume (int): Volume of the media source.
            metadata (Metadata): Metadata on the media.
        """
        self.prev = None
        self.next = None
        self.playlist = None

        self.source = source
        self.duration = duration
        self.volume = volume
        self.metadata = metadata
        self.play = PlayState()
        self.play_queue_state = PlaylistItem.QueueState.OKAY

    def get_playlist(self) -> Playlist:
        """
        Get the playlist of this item.

        Returns:
            Playlist: The associated playlist.
        """
        return self.playlist

    def set_playlist(self, playlist: Playlist):
        """
        Set the playlist of this item.

        Args:
            Playlist: The associated playlist.
        """
        self.playlist = playlist

    def get_position(self) -> int:
        """
        Calculate the position in the playlist.

        Returns:
            int: number in the list of items.
        """
        if self.prev:
            return 1 + self.prev.get_position()
        return 1

    def get_start(self) -> float:
        """
        Calculate the planned start time of an playlist item as an UNIX epoch timestamp.

        The start is the sum of the timeslot start plus all durations of previous items.

        Attention: This should not be mixed up with `self.play.get_play_start()` which represents
        the actual start time. The two are not guaranteed not match!

        Returns:
            float: Timestamp when item is expected to start.
        """
        if self.prev:
            return self.get_prev().get_end()
        else:
            return self.get_playlist().get_timeslot().get_start()

    def get_end(self) -> float:
        """
        Calculate the end of the item as UNIX epoch timestamp.

        Returns:
            float: Timestamp when item is expected to end.
        """
        return self.get_start() + self.duration

    def get_content_type(self) -> resources.ResourceType:
        """
        Get content type for the given source URI.

        Returns:
            ResourceType: Type of the content.
        """
        return resources.ResourceUtil.get_content_type(self.source)

    def get_prev(self) -> PlaylistItem:
        """
        Get previous playlist item.

        Returns:
            PlaylistItem: The previous item.
        """
        return self.prev

    def get_next(self) -> PlaylistItem:
        """
        Get next playlist item.

        Returns:
            PlaylistItem: The next item.
        """
        return self.next

    def get_all_prev(self) -> list[PlaylistItem]:
        """
        Retrieve all previous items as part of the current item's playlist.

        Returns:
            list[PlaylistItem]: All previous items.
        """
        prev = self.get_prev()
        if prev:
            all_prev = prev.get_all_prev()
            all_prev.append(prev)
            return all_prev
        return []

    def get_all_next(self, timeslot_sensitive: bool = True) -> list[PlaylistItem]:
        """
        Retrieve all following items as part of the current item's playlist.

        Args:
            timeslot_sensitive (Boolean): If `True` items which start after \
                the end of the timeslot are excluded.

        Returns:
            list[PlaylistItem]: List of PlaylistItem.

        """
        next = self.get_next()

        if next:
            if timeslot_sensitive:
                if next.get_start() > self.get_playlist().get_timeslot().get_end():
                    return []
            all_next = next.get_all_next(timeslot_sensitive)
            all_next.insert(0, next)
            return all_next
        else:
            return []

    def remove(self):
        """
        Delete the item from the playlist.
        """
        prev = self.prev
        if self.prev:
            self.prev.next = self.next
        if self.next:
            self.next.prev = prev
        self.get_playlist().items.remove(self)
        self.set_playlist(None)
        self.prev = None
        self.next = None

    def __str__(self):
        """
        String representation.

        Returns:
            str: Playlist item as String.

        """
        start = SU.fmt_time(self.get_start())
        end = SU.fmt_time(self.get_end())
        track = self.source[-25:]
        duration = "{:.2f}".format(self.duration)
        return f"PlaylistItem [{start} - {end} " + f"| {duration}sec | Source: ...{track}]"
